import uuid
import xml.etree.ElementTree as ET

from fastapi import Response, HTTPException
from fastapi.datastructures import QueryParams
from sqlmodel import select
import libvirt

from openec2.api.attach_volume import AttachVolumeResponse
from openec2.libvirt import LibvirtSingleton
from openec2.utils.libvirt import instance_to_libvirt_xml
from openec2.db.user import User
from openec2.config import OpenEC2Config
from openec2.db import DatabaseDep
from openec2.db.instance import Instance, EBSVolume
from openec2.utils.libvirt import ebs_volume_to_libvirt_xml


def attach_volume(
    params: QueryParams,
    config: OpenEC2Config,
    db: DatabaseDep,
    user: User,
):
    device = params["Device"]
    instance_id = params["InstanceId"]
    volume_id = params["VolumeId"]

    volume = db.exec(select(EBSVolume).where(EBSVolume.id == volume_id, Instance.owner_id == user.id)).first()
    if volume is None:
        return
    instance = db.exec(select(Instance).where(Instance.id == instance_id, Instance.owner_id == user.id, Instance.terminated == False)).first()
    if instance is None:
        return

    attached_volume_ids = [i.id for i in instance.ebs_volumes]
    if volume_id in attached_volume_ids:
        print("CANNOT ATTACH THE SAME VOLUME TO THE EC2 TWICE")
        return

    if not volume.multi_attach_enabled and volume.instances:
        print("CANNOT ATTACH NON-MULTIATTACH again")
        return

    # Add the required data to libvirt
    conn = LibvirtSingleton.of().connection
    domain = conn.lookupByName(instance_id)

    # Add the memory backing if required
    running = domain.isActive()
    if not instance.ebs_volumes:
        if running:
            raise HTTPException(
                status_code=500,
                detail="Instance is running",
            )

        # Update the instance
        volume.instances.append(instance)

        domain_xml = domain.XMLDesc()
        domain_uuid = ET.fromstring(domain_xml).find("uuid").text

        print(f"Updating XML for {instance.id} with {domain_uuid}")
        instance_xml = instance_to_libvirt_xml(instance, config, domain_uuid)
        print(instance_xml)
        conn.defineXML(instance_xml)
    else:
        # Attach the device
        volume.instances.append(instance)
        domain.attachDeviceFlags(
            ebs_volume_to_libvirt_xml(volume, config),
            libvirt.VIR_DOMAIN_DEVICE_MODIFY_LIVE
            if running
            else libvirt.VIR_DOMAIN_DEVICE_MODIFY_CONFIG,
        )

    db.add(volume)
    db.commit()

    return Response(
        AttachVolumeResponse(
            requestId=uuid.uuid4().hex,
            volumeId=volume_id,
            instanceId=instance_id,
            device=device,
            status="attached",
        ).to_xml(),
        media_type="application/xml",
    )
