import uuid
from typing import cast
from dataclasses import dataclass

from fastapi import Response
from fastapi.datastructures import QueryParams
from sqlmodel import select

from openec2.config import OpenEC2Config
from openec2.db import DatabaseDep
from openec2.db.user import User
from openec2.db.image import AMI
from openec2.api.describe_images import DescribeImagesResponse, Image
from openec2.api.shared import Tag
from openec2.utils.array import parse_array_objects, parse_array_plain


@dataclass
class Filter:
    name: str
    values: list[str]

    def match(self, image: AMI) -> bool:
        value: str | None
        if self.name.startswith("tag:"):
            value = image.tags.get(self.name.replace("tag:", ""))
        else:
            raise Exception(f"Unknown filter name {self.name}")

        return value in self.values


def describe_images(
    params: QueryParams,
    config: OpenEC2Config,
    db: DatabaseDep,
    user: User,
):
    filters: list[Filter] = []
    for filter in parse_array_objects("Filter", cast(dict, params)):
        filters.append(
            Filter(
                name=filter["Name"],
                values=parse_array_plain("Value", filter),
            )
        )

    images: list[Image] = []
    for ami in db.exec(select(AMI).where(AMI.owner_id == user.id)).all():
        if not all(f.match(ami) for f in filters):
            continue

        images.append(
            Image(
                imageId=ami.id,
                imageState="available",
                name=ami.originalFilename,
                tagSet=[
                    Tag(
                        key=key,
                        value=value,
                    ) for key, value in ami.tags.items()
                ],
            ),
        )

    return Response(
        DescribeImagesResponse(
            requestId=uuid.uuid4().hex,
            imagesSet=images,
        ).to_xml(),
        media_type="application/xml",
    )
