import uuid
from typing import cast
import datetime

from fastapi import Response
from fastapi.datastructures import QueryParams
from sqlmodel import select, or_

from openec2.libvirt import LibvirtSingleton
from openec2.api.describe_instances import (
    InstanceDescription,
    DescribeInstancesResponse,
    ReservationSet,
    InstanceState,
    describe_instance,
)
from openec2.api.shared import Tag
from openec2.db.user import User
from openec2.config import OpenEC2Config
from openec2.db import DatabaseDep
from openec2.db.instance import Instance
from openec2.utils.array import parse_array_plain


def describe_instances(
    params: QueryParams,
    config: OpenEC2Config,
    db: DatabaseDep,
    user: User,
):
    if "InstanceId.1" in params:
        instance_ids = parse_array_plain("InstanceId", cast(dict, params))
        instance_expr = [Instance.id == instance_id for instance_id in instance_ids]
        instances = db.exec(
            select(Instance).where(Instance.owner_id == user.id, or_(*instance_expr)),
        ).all()
    else:
        instances = db.exec(
            select(Instance).where(Instance.owner_id == user.id),
        ).all()

    response_items: list[InstanceDescription] = []
    conn = LibvirtSingleton.of().connection
    now = datetime.datetime.now()
    for instance in instances:
        # Include terminated instances for an hour
        if instance.terminated:
            assert instance.terminationDate is not None
            if now <= instance.terminationDate + datetime.timedelta(hours=1):
                response_items.append(
                    InstanceDescription(
                        instanceId=instance.id,
                        imageId=instance.imageId,
                        instanceState=InstanceState(
                            code=48,
                            name="terminated",
                        ),
                        privateIpAddress=instance.privateIPv4,
                        tagSet=[
                            Tag(
                                key=key,
                                value=value,
                            )
                            for key, value in instance.tags.items()
                        ],
                    ),
                )
            continue

        dom = conn.lookupByName(instance.id)
        response_items.append(
            describe_instance(instance, dom),
        )

    return Response(
        DescribeInstancesResponse(
            requestId=uuid.uuid4().hex,
            reservationSet=[
                ReservationSet(
                    reservationId=instance.instanceId,
                    instancesSet=[instance],
                )
                for instance in response_items
            ],
        ).to_xml(),
        media_type="application/xml",
    )
