from typing import Literal, cast
from dataclasses import dataclass
import uuid

from fastapi import Response
from fastapi.datastructures import QueryParams
from sqlmodel import select

from openec2.config import OpenEC2Config
from openec2.db import DatabaseDep
from openec2.db.user import User
from openec2.db.instance import Instance
from openec2.api.describe_tags import DescribeTagsResponse, TagDescription
from openec2.utils.array import parse_array_objects, parse_array_plain, find


@dataclass
class Filter:
    name: Literal["resource-id"] | Literal["resource-type"] | Literal["key"] | str
    values: list[str]

    def match(self, instance: Instance) -> bool:
        if self.name not in instance.tags:
            return False

        value: str | None
        if self.name == "resource-type":
            value = "instance"
        elif self.name == "resource-id":
            value = instance.id
        elif self.name == "key":
            return any(key in instance.tags for key in self.values)
        elif self.name.startswith("tag:"):
            value = instance.tags.get(self.name.replace("tag:", ""))
        else:
            raise Exception(f"Unknown filter name {self.name}")

        return value in self.values

    def value(self, instance: Instance) -> TagDescription:
        if self.name == "resource-type":
            return TagDescription(resourceType="instance")
        elif self.name == "resource-id":
            return TagDescription(resourceId=instance.id)
        elif self.name == "key":
            key_name = find(
                self.values,
                lambda x: x in instance.tags,
            )
            assert key_name is not None

            key_value = instance.tags[key_name]

            return TagDescription(
                key=key_name,
                value=key_value,
            )
        elif self.name.startswith("tag:"):
            tag = self.name.replace("tag:", "")
            return TagDescription(
                key=tag,
                value=instance.tags[tag],
            )
        else:
            raise Exception(f"Unknown filter name {self.name}")

def describe_tags(
    params: QueryParams,
    config: OpenEC2Config,
    db: DatabaseDep,
    user: User,
):
    filters: list[Filter] = []
    for filter in parse_array_objects("Filter", cast(dict, params)):
        filters.append(
            Filter(
                name=filter["Name"],
                values=parse_array_plain("Values", filter),
            )
        )

    response: list[TagDescription] = []
    for instance in db.exec(select(Instance).where(Instance.owner_id == user.id)).all():
        filter = find(
            filters,
            lambda f: f.match(instance),
        )
        if filter is None:
            continue
        response.append(filter.value(instance))

    return Response(
        DescribeTagsResponse(
            requestId=uuid.uuid4().hex,
            nextToken=None,
            tagSet=response,
        ).to_xml(),
        media_type="application/xml",
    )
