import logging
from typing import cast
import uuid
import datetime

from fastapi import Response
from fastapi.datastructures import QueryParams
from sqlmodel import select

from openec2.libvirt import LibvirtSingleton
from openec2.config import OpenEC2Config
from openec2.db import DatabaseDep
from openec2.db.instance import Instance
from openec2.db.user import User
from openec2.utils.array import parse_array_plain
from openec2.images import garbage_collect_image
from openec2.ipam import remove_instance_dhcp_mapping
from openec2.api.shared import InstanceInfo
from openec2.api.describe_instances import describe_instance_state
from openec2.api.terminate_instances import TerminateInstancesResponse, InstancesSet


logger = logging.getLogger()


def terminate_instances(
    params: QueryParams,
    config: OpenEC2Config,
    db: DatabaseDep,
    user: User,
):
    instances: list[InstanceInfo] = []
    conn = LibvirtSingleton.of().connection
    image_ids: set[str] = set()
    for instance_id in parse_array_plain("InstanceId", params):
        instance = db.exec(select(Instance).where(Instance.id == instance_id, Instance.owner_id == user.id)).first()
        if instance is None:
            continue
            # raise HTTPException(status_code=404, detail="Unknown instance")

        dom = conn.lookupByName(instance_id)
        prev_state = describe_instance_state(dom)
        if dom.isActive():
            dom.shutdown()
        current_state = describe_instance_state(dom)
        dom.undefine()

        instances.append(
            InstanceInfo(
                instanceId=instance.id,
                currentState=current_state,
                previousState=prev_state,
            ),
        )

        # Delete files
        logger.debug(f"Removing {config.instances.location / instance_id}")
        instance_disk = config.instances.location / instance_id
        instance_disk.unlink()

        for volume in instance.ebs_volumes:
            volume.instances.remove(instance)


        image_ids.add(cast(str, instance.imageId))
        remove_instance_dhcp_mapping(
            instance.id, instance.interfaceMac, instance.privateIPv4, db
        )

        # Mark the instance as terminated
        instance.terminated = True
        instance.terminationDate = datetime.datetime.now()

    db.commit()

    # Garbage collect AMIs
    for image_id in image_ids:
        garbage_collect_image(image_id, db)

    return Response(
        TerminateInstancesResponse(
            requestId=uuid.uuid4().hex,
            instancesSet=InstancesSet(
                item=instances,
            ),
        ).to_xml(),
        media_type="application/xml",
    )
