from pathlib import Path
from datetime import datetime

from sqlmodel import SQLModel, Field, JSON, Column, Relationship

from openec2.config import OpenEC2Config


class EBSVolumeInstanceLink(SQLModel, table=True):
    instance_id: str | None = Field(
        default=None, foreign_key="instance.id", primary_key=True
    )
    ebs_volume_id: str | None = Field(
        default=None, foreign_key="ebsvolume.id", primary_key=True
    )


class EBSVolume(SQLModel, table=True):
    id: str = Field(primary_key=True)

    availability_zone: str

    multi_attach_enabled: bool

    instances: list["Instance"] = Relationship(
        back_populates="ebs_volumes", link_model=EBSVolumeInstanceLink
    )

    owner_id: int = Field(foreign_key="user.id")

    def path(self, config: OpenEC2Config) -> Path:
        """Compute the path of the volume on disk."""
        return config.instances.volumes / self.id


class Instance(SQLModel, table=True):
    id: str = Field(default=None, primary_key=True)

    # Tags associated with the VM
    tags: dict = Field(sa_column=Column(JSON), default={})

    instanceType: str

    # ImageID of the used AMI. None only if terminated == True.
    imageId: str

    # Optional user data associated with the VM
    userData: str | None

    # MAC of the network interface
    interfaceMac: str

    # Private IPv4 of the instance
    privateIPv4: str

    # The owner that creatd the resource.
    owner_id: int = Field(foreign_key="user.id")

    # Attached EBS volumes
    ebs_volumes: list[EBSVolume] = Relationship(
        back_populates="instances", link_model=EBSVolumeInstanceLink
    )

    # Is the instance terminated
    terminated: bool

    # Date at which the instance got terminated
    terminationDate: datetime | None
