from urllib.parse import parse_qs
from typing import cast

from fastapi import FastAPI, HTTPException, Request, Response
from sqlalchemy import select
from sqlmodel import SQLModel

from openec2.security.aws import AWSSignature
from openec2.utils.text import multiline_yaml_response
from openec2.config import OpenEC2Config
from openec2.db import DatabaseDep, engine
from openec2.db.user import User
from openec2.actions.describe_images import describe_images
from openec2.actions.import_image import import_image
from openec2.actions.describe_instances import describe_instances
from openec2.actions.run_instances import run_instances
from openec2.actions.terminate_instances import terminate_instances
from openec2.actions.start_instances import start_instances
from openec2.actions.stop_instances import stop_instances
from openec2.actions.deregister_image import deregister_image
from openec2.actions.create_volume import create_volume
from openec2.actions.attach_volume import attach_volume
from openec2.actions.describe_volumes import describe_volumes
from openec2.actions.detach_volume import detach_volume
from openec2.db.instance import Instance
from openec2.actions.get_caller_identity import get_caller_identity
from openec2.actions.describe_instance_types import describe_instance_types
from openec2.actions.describe_tags import describe_tags
from openec2.actions.create_tags import create_tags
from openec2.actions.describe_instance_attribute import describe_instance_attribute


app = FastAPI()


@app.on_event("startup")
def on_startup():
    SQLModel.metadata.create_all(engine)


@app.get("/healthz", response_model=None)
def healthz():
    return {
        "status": "OK",
    }


@app.get("/Action", response_model=None)
def get_action(
    request: Request, config: OpenEC2Config, db: DatabaseDep, user: AWSSignature
):
    print("GET Action")
    return run_action(
        request,
        config,
        db,
        cast(dict, request.query_params),
        user,
    )


@app.post("/Action", response_model=None)
async def post_action(
    request: Request, config: OpenEC2Config, db: DatabaseDep, user: AWSSignature
):
    print("POST Action")

    body = (await request.body()).decode()
    print(f"--> {body}")
    query_params = {
        key: value[0]
        for key, value in parse_qs(body).items()
    }
    return run_action(
        request,
        config,
        db,
        cast(dict, query_params),
        user,
    )


def run_action(
    request: Request,
    config: OpenEC2Config,
    db: DatabaseDep,
    query_params: dict[str, str],
    user: User,
):
    action = query_params["Action"]
    action_function = {
        "ImportImage": import_image,
        "DescribeImages": describe_images,
        "RunInstances": run_instances,
        "DescribeInstances": describe_instances,
        "TerminateInstances": terminate_instances,
        "StartInstances": start_instances,
        "StopInstances": stop_instances,
        "DeregisterImage": deregister_image,
        "CreateVolume": create_volume,
        "AttachVolume": attach_volume,
        "DescribeVolumes": describe_volumes,
        "DetachVolume": detach_volume,
        "GetCallerIdentity": get_caller_identity,
        "DescribeInstanceTypes": describe_instance_types,
        "DescribeTags": describe_tags,
        "CreateTags": create_tags,
        "DescribeInstanceAttribute": describe_instance_attribute,
    }.get(action)
    if action_function is None:
        print(f"Unknown action: '{action}'")
        raise HTTPException(
            status_code=404,
            detail="Unknown action",
        )

    return action_function(query_params, config, db, user)



@app.get("/private/cloudinit/{instance_id}/{entry}")
def cloud_init_data(instance_id: str, entry: str, db: DatabaseDep):
    if entry not in ("user-data", "meta-data", "vendor-data", "network-config"):
        raise HTTPException(status_code=404, detail="Unknown cloud-init file")

    if entry == "user-data":
        instance = db.exec(select(Instance).where(Instance.id == instance_id)).first()[
            0
        ]
        if instance is None:
            raise HTTPException(status_code=404, detail="Unknown instance")

        if instance.userData is None:
            raise HTTPException(status_code=404, detail="No user data")
        return Response(
            content=instance.userData,
            media_type="application/yaml",
        )
    elif entry == "meta-data":
        return multiline_yaml_response(
            [
                f"instance-id: {instance_id}",
                f"local-hostname: {instance_id}",
            ]
        )
    elif entry == "vendor-data":
        return multiline_yaml_response(
            [
                "#cloud-config",
                "growpart:",
                "   devices: [/]",
                "   ignore_growroot_disabled: true",
            ]
        )
    elif entry == "network-config":
        return Response(
            "",
            media_type="application/yaml",
        )
